-- Proof that there are an infinite number of primes.
-- Author : Thierry Coquand
-- Created: 2005-05-12

-- # import Property

data Nat : Set where
 zero : Nat
 succ : Nat -> Nat

(+) : Nat -> Nat -> Nat
x + zero = x
x + succ y = succ (x + y)

lem : (x:Nat) -> zero + x == x
lem zero = fol-plugin()
lem (succ x) = fol-plugin(lem x)

lem1 : (x,y:Nat) -> (succ x) + y == succ (x+y)
lem1 x zero = fol-plugin()
lem1 x (succ y) = fol-plugin(lem1 x y)

comAdd :  (x,y:Nat) -> x + y == y + x
comAdd x zero = fol-plugin(lem)
comAdd x (succ y) = fol-plugin(comAdd x y,lem1)

assocAdd :  (x,y,z:Nat) -> (x + y) + z == x + (y + z)
assocAdd x y zero = fol-plugin()
assocAdd x y (succ z) = fol-plugin(assocAdd x y z)


(*) : Nat -> Nat -> Nat
x * zero = zero
x * succ y = x + (x * y)

lem2 : (x:Nat) -> zero * x == zero
lem2 zero = fol-plugin()
lem2 (succ x) = fol-plugin(lem2 x,lem)

lem3 : (x,y:Nat) -> (succ x) * y == (x * y) + y
lem3 x zero = fol-plugin()
lem3 x (succ y) = fol-plugin(lem3 x y,lem1,assocAdd)

comMul : (x,y:Nat) -> x * y == y * x
comMul x zero = fol-plugin(lem2)
comMul x (succ y) = fol-plugin(comMul x y,lem3,assocAdd,comAdd)

distr : (x,y,z:Nat) -> x * (y + z) == (x * y) + (x * z)
distr x y zero = fol-plugin()
distr x y (succ z) = fol-plugin(distr x y z,assocAdd,comAdd)

assocMul : (x,y,z:Nat) -> x * (y * z) == (x * y) * z
assocMul x y zero = fol-plugin()
assocMul x y (succ z) = fol-plugin(assocMul x y z,distr)

(|) : Nat -> Nat -> Prop

axDiv : (x,y:Nat) -> x | y <=> Exists (\ z -> y == x * z)

lem4 : (x:Nat) -> x | zero
lem4 x = fol-plugin((*),axDiv)

lem5 : (x,y,z:Nat) -> x | y /\ y | z ==> x | z
lem5 x y z = fol-plugin(assocMul,axDiv)

(<=) : Nat -> Nat -> Prop

axLeq : (x,y:Nat) -> x <= y <=> Exists (\ z -> y == x + z)

lem6 : (x,y,z:Nat) -> x <= y /\ y <= z ==> x <= z
lem6 x y z = fol-plugin(assocAdd,axLeq)

(<) : Nat -> Nat -> Prop

axLess : (x,y:Nat) -> x < y <=> Exists (\ z -> y == succ (x + z))

lem7 : (x,y:Nat) -> x < y <=> succ x <= y
lem7 x y = fol-plugin(axLess,axLeq,lem1,(+))

lem8 : (x,y,z:Nat) -> x <= y /\ y < z ==> x < z
lem8 x y z = fol-plugin((+),assocAdd,axLeq,axLess)

Prime : Nat -> Prop

one : Nat 
one = succ zero

axPrime : (x:Nat) -> Prime x <=> ((y:Nat) -> one < y /\ y < x ==> Not (y|x))

lem9 : (x:Nat) -> x * one == x
lem9 x = fol-plugin()

fact : Nat -> Nat
fact zero = one
fact (succ x) = (succ x) * (fact x)

caseNat : (x:Nat) -> x == 0 \/ Exists (\ z -> x == succ z)
caseNat zero = fol-plugin()
caseNat (succ x) = fol-plugin()

lem10 : (x,y:Nat) -> x <= succ y ==> x == succ y \/ x <= y
lem10 x y = impI (\ hyp -> fol-plugin(axLeq,hyp,caseNat,Nat,(+)))

lem11 : (x,y:Nat) -> x | fact y ==> x | fact (succ y)
lem11 x y =
 let 
   rem : (z:Nat) -> fact z | fact (succ z)
   rem z = fol-plugin(comMul,axDiv)
 in
  fol-plugin(fact,lem5,rem)

lem12 : (x,y:Nat) -> one <= y /\ y <= x ==> y | fact x
lem12 zero y = fol-plugin(axLeq,Nat,lem1)
lem12 (succ x) y =
 let 
  rem : (succ x) | fact (succ x)
  rem = fol-plugin(axDiv)
 in
  fol-plugin(lem12 x y,lem10,lem11,fact,rem)

lem13 : (x,y:Nat) -> x <= y <=> succ x <= succ y
lem13 x y = fol-plugin(Nat,axLeq,lem1)

lem14 : (x,y:Nat) -> x < y <=> succ x < succ y
lem14 x y = fol-plugin(Nat,axLess,lem1)

lem15 : (x,y:Nat) -> x <= y \/ y < x
lem15 zero zero = fol-plugin(axLeq,axLess,(+))
lem15 zero (succ y) = fol-plugin(axLeq,axLess,(+),lem)
lem15 (succ x) zero = fol-plugin(axLeq,axLess,(+),lem)
lem15 (succ x) (succ y) = fol-plugin(lem15 x y,lem13,lem14)

lem16 : (x:Nat) -> Not (x<x)
lem16 zero = fol-plugin(Nat,(+),axLess)
lem16 (succ x) = fol-plugin(lem16 x,lem14)

lem17 : (x,y,z:Nat) -> x < y /\ y <= z ==> x < z
lem17 x y z = fol-plugin((+),assocAdd,axLeq,axLess,lem1)

lem18 : (x,y:Nat) -> x < y <=> Not (y <= x)
lem18 x y = fol-plugin(lem8,lem16,lem17,lem15)

lem19 : (x,y,z:Nat) -> y <= z ==> x * y <= (x * z)
lem19 x y z = fol-plugin(axLeq,distr)

lem20 : (x,y,z:Nat) -> y < z ==> (x * y) + x <= (x * z)
lem20 x y z = fol-plugin(lem7,lem19,(*),comAdd)

lem21 : (x,y,z:Nat) -> x + y == x + z ==> y == z
lem21 zero y z = fol-plugin(lem)
lem21 (succ x) y z = fol-plugin(lem21 x y z,lem1,Nat)

lem22 : (x,y,z:Nat) -> x + y <= (x + z) ==> y <= z
lem22 x y z = fol-plugin(lem21,axLeq,assocAdd)

lem23 : (x:Nat) -> x <= one ==> x == zero \/ x == one
lem23 zero = fol-plugin()
lem23 (succ zero) = fol-plugin()
lem23 (succ (succ x)) = fol-plugin(axLeq,Nat,lem1)

prop1 : (x,y:Nat) -> x | y /\ x | succ y ==> x == one
prop1 x y =
 let 
  rem : (t,u:Nat) -> y == x * t /\ succ y == x * u ==> x == one
  rem t u = impI (\ hyp ->
                     let 
                       rem1 : t < u
                       rem1 = fol-plugin(hyp,lem15,lem19,lem16,lem7)
                       rem2 : y + x <= (y + one)
                       rem2 = fol-plugin(hyp,lem20,(+),rem1)
                       rem3 : x <= one
                       rem3 = fol-plugin(rem2,lem22,(+))
                     in
                      fol-plugin(rem3,hyp,lem23,Nat,lem2))
 in
  fol-plugin(rem,axDiv)

Main : Nat -> Nat -> Prop
Main a m = Exists (\ z -> z <= m /\ Prime z /\ a < z) \/
           ((y:Nat) -> one < y /\ y <= m ==> Not (y | succ (fact a)))


lem24 : (x,y:Nat) -> x < succ y <=> x <= y
lem24 x y = fol-plugin(axLess,axLeq,lem1,Nat)


prop2 : (a,m:Nat) -> Main a m
prop2 a zero = 
 let
  rem : (y:Nat) -> one < y ==> Not (y <= zero)
  rem y = fol-plugin(lem17,axLess,Nat)
 in 
  fol-plugin(rem)
prop2 a (succ m) =
 let
  indHyp : Main a m
  indHyp = prop2 a m
  case1 : (z:Nat) -> z <= m /\ Prime z /\ a < z ==> Main a (succ m)
  case1 z = impI (\ hyp -> let 
                             rem : z <= succ m
                             rem = fol-plugin(hyp,(+),axLeq)
                            in
                              fol-plugin(rem,hyp))
  case2 : ((y:Nat) -> one < y /\ y <= m ==> Not (y | succ (fact a))) ==> Main a (succ m)
  case2 = 
   impI ( \ hyp ->
            let
              case21 : succ m | succ (fact a) ==> Main a (succ m)
              case21 = impI ( \ hyp1 ->
                                let
                                  rem1 : Prime (succ m)
                                  rem1 =  let
                                            rem11 : (y:Nat) -> one < y /\ y < succ m ==> Not (y | succ m)
                                            rem11 y = impI( \ hyp2 -> fol-plugin(hyp2,hyp,hyp1,lem5,lem24))
                                          in
                                           fol-plugin(axPrime,rem11)
                                  rem2 : a < succ m \/ m == zero
                                  rem2 = let
                                           rem21 : (x:Nat) -> one <= x /\ x <= a /\ x|succ (fact a) ==> x == one
                                           rem21 x = impI (\ hyp2 -> fol-plugin(lem12,hyp2,prop1))
                                           rem22 : one <= succ m
                                           rem22 = fol-plugin(lem,axLeq,lem1)
                                         in
                                          fol-plugin(lem15,rem21,rem22,hyp1,Nat)
                                  rem3 : succ m <= succ m
                                  rem3 = fol-plugin(axLeq,(+))
                                  rem4 : (y:Nat) -> one < y /\ y <= one ==> Not (y | succ (fact a))
                                  rem4 y = fol-plugin(lem18)
                                in
                                 fol-plugin(rem1,rem2,rem3,rem4)
                             )
              case22 : (Not (succ m | succ (fact a))) ==> Main a (succ m)
              case22 = impI ( \ hyp1 ->
                                let
                                 rem1 : (d:Nat) -> one < d /\ d <= succ m ==> Not (d | succ (fact a))
                                 rem1 d = impI (\ hyp2 -> 
                                                   let
                                                    rem11 : d <= m \/ d == succ m
                                                    rem11 = fol-plugin(lem10,hyp2)
                                                    rem12 : d <= m ==> Not (d | succ (fact a))
                                                    rem12 = impI (\ hyp3 -> fol-plugin(hyp3,hyp2,hyp))
                                                   in
                                                    fol-plugin(rem11,rem12,hyp1)
                                                )
                                in 
                                 fol-plugin(rem1)
                             )
             in
              fol-plugin(case21,case22)
         )
 in
  fol-plugin(indHyp,case1,case2)


-- infinity of primes

witness (P:Nat-> Prop) : (x:Nat) -> P x -> Exists P
witness x h = fol-plugin(h)

lem25 : (x:Nat) -> x | x
lem25 x =
 let
  rem : Exists (\ z -> x == x * z)
  rem = witness one fol-plugin()
 in
  fol-plugin(axDiv,rem)

lem26 : (x:Nat) -> x <= x
lem26 x =
 let
  rem : Exists (\ z -> x == x + z)
  rem = witness zero fol-plugin()
 in
  fol-plugin(axLeq,rem)

lem27 : (x:Nat) -> zero < succ x
lem27 x = fol-plugin(axLess,lem)

lem28 : (x,y:Nat) -> zero < x /\ zero < y ==> zero < (x * y)
lem28 zero zero = fol-plugin(lem16)
lem28 zero (succ y) = fol-plugin(lem16)
lem28 (succ x) zero = fol-plugin(lem16)
lem28 (succ x) (succ y) = fol-plugin(lem27,(*),lem1)

lem30 : (x:Nat) -> zero < fact x
lem30 zero = fol-plugin(lem27)
lem30 (succ x) = fol-plugin(lem30 x,lem28,lem27)


thm : (a:Nat) -> Exists (\ z -> Prime z /\ a < z)
thm a = 
  fol-plugin(lem25,lem26,lem30,lem14,prop2 a (succ (fact a)))




