
-- # import QuickCheck

-- Equality ---------------------------------------------------------------

-- We can't prove this in the system. Annoying.
subst : (A:Set) -> (B : A -> Type) -> (x, y : A) -> Eq A x y ->
	  B x -> B y

-- Bool -------------------------------------------------------------------

fancyIf : (A:Type) -> (x : Bool) ->
	    (Holds x -> A) -> (Holds (not x) -> A) -> A
fancyIf A true f g  = f tt
fancyIf A false f g = g tt

-- Nat --------------------------------------------------------------------

-- List -------------------------------------------------------------------

append : (A:Set) -> List A -> List A -> List A
append A [] ys	    = ys
append A (x::xs) ys  = x :: append A xs ys

(++) = append Nat

-- Maybe ------------------------------------------------------------------

data Maybe (A:Set) : Set where
  nothing : Maybe A
  just	  : A -> Maybe A

arbMaybe : (A:Set) -> Gen A -> Gen (Maybe A)
arbMaybe A g = frequency  [ (1,  return nothing)
			  , (10, liftM A (Maybe A) just g)
			  ]

maybe : (A, B:Set) -> B -> (A -> B) -> Maybe A -> B
maybe A B n j nothing	= n
maybe A B n j (just x)	= j x

-- Ranges -----------------------------------------------------------------

Range = { Maybe Nat ; Maybe Nat }

arbRange : Gen Range
arbRange = arbSigma (Maybe Nat) (\_ -> Maybe Nat)
	      (arbMaybe Nat arbNat)
	      (\_ -> arbMaybe Nat arbNat)

infix 10 `between`
between : Nat -> Range -> Bool
n `between` (lo, hi) =
  maybe Nat Bool true (\lo' -> lo' <= n) lo &&
  maybe Nat Bool true (\hi' -> n <= hi') hi

lower : Nat -> Range -> Range
lower n (lo, hi) = (lo, just n)

upper : Nat -> Range -> Range
upper n (lo, hi) = (just n, hi)

expand : Nat -> Range -> Range
expand n (lo, hi) =
  ( just (maybe Nat Nat n (min n) lo)
  , just (maybe Nat Nat n (max n) hi)
  )

validRange : Range -> Bool
validRange (lo,hi) =
  let r : Range
      r = (lo, hi)
  in  maybe Nat Bool true (\n -> n `between` r) lo &&
      maybe Nat Bool true (\n -> n `between` r) hi

PropBetweenExpand = (n:Nat) -> (r:Range) ->
    Holds (validRange r) ==>
    Holds (n `between` expand n r)

--testBetweenExpand : PropBetweenExpand
--testBetweenExpand = qc-plugin(arbNat, arbRange)

propBetweenExpand : PropBetweenExpand
propBetweenExpand n (nothing, nothing)	= fol-plugin( propLeqRefl, propAnd' )
propBetweenExpand n (nothing, just hi)	= fol-plugin( propLeqRefl
						    , propMaxLeq
						    , propAnd'
						    )
propBetweenExpand n (just lo, nothing)	= fol-plugin( propLeqRefl
						    , propMinLeq
						    , propAnd'
						    )
propBetweenExpand n (just lo, just hi)	= fol-plugin( propMinLeq
						    , propMaxLeq
						    , propAnd'
						    )

PropLowerExpand = (n,m:Nat) -> (r:Range) ->
		  Holds (validRange r)	==>
		  Holds (m `between` r) ==>
		  Holds (n <= m)	==>
		  Eq Range  (expand n (lower m r))
			    (lower m (expand n r))

--testLowerExpand : PropLowerExpand
--testLowerExpand = qc-plugin(arbNat, arbNat, arbRange)

propLowerExpand : PropLowerExpand
propLowerExpand n m (nothing, nothing) = fol-plugin( propMaxEq )
propLowerExpand n m (just lo, nothing) = fol-plugin( propMaxEq )
propLowerExpand n m (nothing, just hi) = fol-plugin( propMaxEq )
propLowerExpand n m (just lo, just hi) = fol-plugin( propMaxEq )

-- Binary search trees ----------------------------------------------------

data BST (r : Range) : Type where
  leaf	: BST r
  node	: (n:Nat) -> Holds (n `between` r) ->
	    BST (lower n r) -> BST (upper n r) -> BST r

empty : BST (nothing, nothing)
empty = leaf

flatten : (r:Range) -> BST r -> List Nat
flatten r leaf			= []
flatten r (node n _ left right)	=
  flatten (lower n r) left ++ [n] ++ flatten (upper n r) right

insert : (r:Range) -> Holds (validRange r) ->
	  (n:Nat) -> BST r -> BST (expand n r)
insert r rValid n leaf =
  node n (propBetweenExpand n r rValid) leaf leaf
insert r rValid n (node m btw left right) =
  fancyIf (BST (expand n r)) (n <= m)
	  (\p -> node m assert-plugin()
		    ( subst Range BST
			(expand n (lower m r))
			(lower m (expand n r))
			(propLowerExpand n m r rValid btw p)
			(insert (lower m r) assert-plugin() n left)
		    )
		    ( subst Range BST
			(upper m r)
			(upper m (expand n r))
			assert-plugin()
			right
		    )
	  )
	  (\q -> node m assert-plugin()
		    ( subst Range BST
			(lower m r)
			(lower m (expand n r))
			assert-plugin()
			left
		    )
		    ( subst Range BST
			(expand n (upper m r))
			(upper m (expand n r))
			assert-plugin()
			(insert (upper m r) assert-plugin() n right)
		    )
	  )

