
-- # import Rel
-- # import Setoid
-- # import ChainReasoning

-- Preorder ---------------------------------------------------------------

sig Preorder (A:SET) : Type
  extends ReflexiveRel
	, TransitiveRel where

-- Partial order ----------------------------------------------------------

sig PartialOrder (A:Setoid) : Type
  extends Preorder
	, AntiSymmetricRel
	, ReflexiveSetoidRel where

-- Total order ------------------------------------------------------------

sig TotalOrder (A:Setoid) : Type
    extends PartialOrder, TotalRel where

-- Strict partial order ---------------------------------------------------

sig StrictPartialOrder (A:Setoid) : Type
  extends IrreflexiveRel
	, AsymmetricRel
	, TransitiveRel
	, SetoidRel where

mkStrictPartialOrder (A:Setoid) : (R:Rel A) ->
  Prf (Irreflexive R) -> Prf (Asymmetric R) ->
  Prf (Transitive R) -> Prf (RespectsEq R) ->
  StrictPartialOrder A
mkStrictPartialOrder R irrefl asym trans respeq =
  struct rel	= R.rel
	 irrefl = irrefl
	 asym	= asym
	 trans	= trans
	 respeq = respeq

-- Strict total order -----------------------------------------------------

data Trichotomy' {A:Setoid}(R:Rel A)(x,y:El A) : Prop where
  triLess    : Prf (R.rel x y) -> Trichotomy' R x y
  triEq	     : Prf (x == y)    -> Trichotomy' R x y
  triGreater : Prf (R.rel y x) -> Trichotomy' R x y

Trichotomy (A:Setoid) : Rel A -> El A -> El A -> PROP
Trichotomy R x y = mkPROP (Trichotomy' R x y)

Trichotomous (A:Setoid) : Rel A -> PROP
Trichotomous R = ForAll $ \x -> ForAll $ \y -> Trichotomy R x y

sig StrictTotalOrder (A:Setoid) : Type
    extends StrictPartialOrder where
  trich : Prf (Trichotomous (mkRel rel))

trichotomy (A:Setoid)(R:StrictTotalOrder A)(x,y:El A) :
  Prf (Trichotomy R x y)
trichotomy {_} {R} = R.trich _ _

-- Every partial order induces a strict partial order ---------------------

strictPartialOrder (A:Setoid) : PartialOrder A -> StrictPartialOrder A
strictPartialOrder {A} po =
  struct rel	= (<)
	 irrefl = irrefl
	 asym	= asym
	 trans	= trans
	 respeq = respeq
  where
    infix 10 <=, <

    (<=) = po.rel

    (<) : El A -> El A -> PROP
    x < y = x <= y /\ Not (x == y)

    less = mkRel (<)

    respeq : Prf (RespectsEq less)
    respeq x y z w "x=y" "z=w" (andI "x<=z" "x!=z") =
      andI (respectsEq "x=y" "z=w" "x<=z") "y!=w"
      where
	"y!=w" : Prf (Not (y == w))
	"y!=w" "y=w" =
	  notE "x!=z"
	  $ chainReasoning
	  $ x >>>> y  `since` "x=y"
	       >>> w  `since` "y=w"
	       >>> z  `since` symmetry "z=w"

    irrefl : Prf (Irreflexive less)
    irrefl x (andI _ "x!=x") = notE "x!=x" (reflexivity {A})

    asym : Prf (Asymmetric less)
    asym x y (andI "x<=y" "x!=y") (andI "y<=x" "y!=x") =
      notE "x!=y" (antisymmetry "x<=y" "y<=x")

    trans : Prf (Transitive less)
    trans x y z (andI "x<=y" "x!=y")
		(andI "y<=z" "y!=z")
		= andI "x<=z" "x!=z"
      where
	"x<=z" = transitivity {A} "x<=y" "y<=z"
	"x!=z" : Prf (Not (x == z))
	"x!=z" "x=z" = notE "x!=y" $ antisymmetry "x<=y" "y<=x"
	  where
	    "z<=x" : Prf (z <= x)
	    "z<=x" = setoidRefl (symmetry "x=z")
	    "y<=x" = transitivity {A} "y<=z" "z<=x"

-- And every strict partial order induces a non-strict partial order ------

nonStrictPartialOrder (A:Setoid) : StrictPartialOrder A -> PartialOrder A
nonStrictPartialOrder {A} spo =
  struct rel	 = (<=)
	 refl	 = refl
	 trans	 = trans
	 antisym = antisym
	 respeq  = respeq
  where
    infix 10 <, <=

    (<) = spo.rel
    
    (<=) : El A -> El A -> PROP
    x <= y = x < y  \/  x == y

    leq = mkRel (<=)

    refl : Prf (Reflexive leq)
    refl x = orI2 reflexivity

    respeq : Prf (RespectsEq leq)
    respeq x y z w "x=y" "z=w" (orI1 "x<z") =
      orI1 $ respectsEq "x=y" "z=w" "x<z"
    respeq x y z w "x=y" "z=w" (orI2 "x=z") =
      orI2 $ chainReasoning
	   $ y	>>>> x	`since` symmetry "x=y"
		 >>> z	`since` "x=z"
		 >>> w	`since` "z=w"

    antisym : Prf (AntiSymmetric leq)
    antisym x y (orI2 "x=y")  _		  = "x=y"
    antisym x y (orI1 _)     (orI2 "y=x") = symmetry "y=x"
    antisym x y (orI1 "x<y") (orI1 "y<x") = falseE $ asymmetry "x<y" "y<x"

    trans : Prf (Transitive leq)
    trans x y z (orI1 "x<y") (orI1 "y<z") =
      orI1 $ transitivity {A} "x<y" "y<z"
    trans x y z (orI1 "x<y") (orI2 "y=z") =
      orI1 $ respectsEq (reflexivity {A}) "y=z" "x<y"
    trans x y z (orI2 "x=y") (orI1 "y<z") =
      orI1 $ respectsEq (symmetry {A} "x=y") (reflexivity {A}) "y<z"
    trans x y z (orI2 "x=y") (orI2 "y=z") =
      orI2 $ transitivity {A} "x=y" "y=z"

