
-- # import Set
-- # import Prop
-- # import Pred

-- Relations --------------------------------------------------------------

sig Rel (A:SET) : Type where
  rel : El A -> El A -> PROP

mkRel (A:SET) : (El A -> El A -> PROP) -> Rel A
mkRel rel = struct rel = rel

-- Reflexive relations ----------------------------------------------------

Reflexive (A:SET) : Rel A -> PROP
Reflexive R = ForAll $ \x -> R.rel x x

sig ReflexiveRel (A:SET) : Type extends Rel where
  refl : Prf (Reflexive (mkRel rel))

reflexivity (A:SET)(R:ReflexiveRel A)(x:El A) : Prf (R.rel x x)
reflexivity {_} {R} = R.refl _

-- Irreflexive relations --------------------------------------------------

Irreflexive (A:SET) : Rel A -> PROP
Irreflexive R = ForAll $ \x -> Not (R.rel x x)

sig IrreflexiveRel (A:SET) : Type extends Rel where
  irrefl : Prf (Irreflexive (mkRel rel))

irreflexivity (A:SET)(R:IrreflexiveRel A)(x:El A) : Prf (Not (R.rel x x))
irreflexivity {_} {R} = R.irrefl _

-- Symmetric relations ----------------------------------------------------

Symmetric (A:SET) : Rel A -> PROP
Symmetric R = ForAll $ \x -> ForAll $ \y ->
	      R.rel x y ==> R.rel y x

sig SymmetricRel (A:SET) : Type extends Rel where
  sym : Prf (Symmetric (mkRel rel))

symmetry (A:SET)(R:SymmetricRel A)(x,y:El A) :
  Prf (R.rel x y ==> R.rel y x)
symmetry {_} {R} xy = R.sym _ _ xy

-- Asymmetric relations ---------------------------------------------------

Asymmetric (A:SET) : Rel A -> PROP
Asymmetric R =
  ForAll $ \x ->
  ForAll $ \y ->
    R.rel x y ==> Not (R.rel y x)

sig AsymmetricRel (A:SET) : Type extends Rel where
  asym : Prf (Asymmetric (mkRel rel))

asymmetry (A:SET)(R:AsymmetricRel A)(x,y:El A) :
  Prf (R.rel x y ==> Not (R.rel y x))
asymmetry {_} {R} xRy = R.asym _ _ xRy

-- Transitive relations ---------------------------------------------------

Transitive (A:SET) : Rel A -> PROP
Transitive R =
  ForAll $ \x ->
  ForAll $ \y ->
  ForAll $ \z ->
    R.rel x y ==> R.rel y z ==> R.rel x z

sig TransitiveRel (A:SET) : Type extends Rel where
  trans : Prf (Transitive (mkRel rel))

transitivity (A:SET)(R:TransitiveRel A)(x,y,z:El A) :
  Prf (R.rel x y ==> R.rel y z ==> R.rel x z)
transitivity {_} {R} xy yz = R.trans _ _ _ xy yz

-- Total relations --------------------------------------------------------

Total (A:SET) : Rel A -> PROP
Total R = ForAll $ \x -> ForAll $ \y ->
	  R.rel x y \/ R.rel y x

-- A total relation is also reflexive so it makes sense to inherit from
-- ReflexiveRel.
sig TotalRel (A:SET) : Type extends ReflexiveRel where
  total : Prf (Total (mkRel rel))

totality (A:SET)(R:TotalRel A)(x,y:El A) :
  Prf (R.rel x y \/ R.rel y x)
totality {_} {R} = R.total _ _

-- Equivalence relations --------------------------------------------------

sig EquivalenceRel (A:SET) : Type
  extends ReflexiveRel
	, SymmetricRel
	, TransitiveRel
  where

-- Substitutive relations -------------------------------------------------

Substitutive (A:SET) : Rel A -> Pred A -> PROP
Substitutive R P =
  ForAll $ \x ->
  ForAll $ \y ->
    R.rel x y ==> P.pred x ==> P.pred y

sig SubstitutiveRel (A:SET) : Type extends Rel where
  subst : (P : Pred A) -> Prf (Substitutive (mkRel rel) P)

-- The empty relation -----------------------------------------------------

sig SymTransSubstRel (A:SET) : Type
  extends SymmetricRel
	, TransitiveRel
	, SubstitutiveRel where

emptyRel (A:SET) : SymTransSubstRel A
emptyRel =
  struct rel	 x y	     = False
	 sym	 x y   xy    = xy
	 trans	 x y z xy yz = xy
	 subst P x y   xy px = falseE xy

-- Relations between relations --------------------------------------------

(===) (A:SET) : Rel A -> Rel A -> PROP
R === S = ForAll $ \x -> ForAll $ \y ->
	  R.rel x y <==> S.rel x y

(=<=) (A:SET) : Rel A -> Rel A -> PROP
R =<= S = ForAll $ \x -> ForAll $ \y ->
	  R.rel x y ==> S.rel x y

