{-| Structural beta-eta-equality and universe-subtyping for values.
-}

module TypeChecker.Equality where

import Control.Monad		    ( unless )

import Syntax.Internal

import TypeChecker.Monad
import TypeChecker.Monad.Errors
import TypeChecker.Monad.Options
import TypeChecker.UnfoldConstants  ( unfoldConstants )
import TypeChecker.Values	    ( vApp )
import TypeChecker.Errors

import Utils.Names		    ( fresh )
import Utils.Monads

-- Helpers ----------------------------------------------------------------

freshValue :: Name -> Type -> TC Value
freshValue x t = VId <$> fresh x <*> return t

-- Equality ---------------------------------------------------------------

(===) :: Value -> Value -> TC ()

v1 === v2 = unfoldConstants v1 ==* unfoldConstants v2

VId "_" _ ==* _	= failure dontCareNotEqualError
_ ==* VId "_" _	= failure dontCareNotEqualError

VId x _ ==* VId y _
    | x == y    = return ()
    | otherwise = failure $ variablesNotEqualError x y

VDef x _ ==* VDef y _
    | x == y    = return ()
    | otherwise = failure $ variablesNotEqualError x y

VLam f t ==* VLam g _ =
    do	let ~(VPi w _ _) = unfoldConstants t
	x <- freshValue (clName g) w
	clApp f x === clApp g x

VLam f t ==* u = 
    do	let ~(VPi w _ _) = unfoldConstants t
	x <- freshValue (clName f) w
	clApp f x === vApp u x

u ==* VLam f t =
    do	let ~(VPi w _ _) = unfoldConstants t
	x <- freshValue (clName f) w
	VApp u x === clApp f x

VPi v f _ ==* VPi w g _ =
    do	v === w
	x <- freshValue (clName g) v
	clApp f x === clApp g x

VSort s1 ==* VSort s2 =
    unless (s1 == s2) $ failure $ sortsNotEqualError s1 s2

VApp u1 v1 ==* VApp u2 v2 = -- an application in normal form
    do	u1 === u2	    -- have the first argument in normal form
	v1 === v2	    -- the second is only in normal form if
			    -- the first is a constant

VTop _ ==* VTop _ = return ()

u ==* v = failure $ notAtAllEqualError u v

data Tree = Leaf Value | Node Tree Tree

-- | Non-empty list. This is used when inferring the type
--   of a multiple clause definition or a case. Only works
--   for a limited class of definitions, of course.
allEqual :: [Value] -> TC Value
allEqual = treeEqual . mkTree . map Leaf
    where
	mkTree [t]	= t
	mkTree (l:r:ts) = mkTree (Node l r : ts)

	treeEqual (Leaf v) = return v
	treeEqual (Node t1 t2) =
	    do	v1 <- treeEqual t1
		v2 <- treeEqual t2
		v1 === v2
		return v1

-- Subtyping --------------------------------------------------------------

(=<=) :: Value -> Value -> TC ()
v =<= w = unfoldConstants v =<=* unfoldConstants w

VSort s1 =<=* VSort s2 =
    unless (s1 <= s2) $ failure $ sortsNotLeqError s1 s2

VDef x _ =<=* VDef y _
    | x == y	= return ()

-- ? Alternative: v === w, theoretical implications ?
VPi v f _ =<=* VPi w g _ =
    do	w =<= v
	x <- freshValue (clName g) w
	clApp f x =<= clApp g x

v =<=* w = v === w

