{-| Reconstructing a term from a value. Useful for plugins
    that wants to generate terms that are not guaranteed to
    be type correct. Does not work for 'VTop'.
-}
module TypeChecker.ValueToTerm where

import Syntax.Internal
import Syntax.Position

import TypeChecker.Monad
import TypeChecker.Monad.Errors
import TypeChecker.UnfoldConstants
import TypeChecker.Errors

import Utils.Names
import Utils.Monads

valueToTerm :: Value -> TC Term
valueToTerm v =
    case v of
	VId x _			-> return $ Var x ()
	VDef x _		-> return $ Def x ()
	VPi v f _		->
	    do	a <- valueToTerm v
		x <- fresh (clName f)
		b <- valueToTerm $ f `clApp` (VId x v)
		return $ Pi x a b ()
	VLam f t		->
	    do	let ~(VPi w _ _) = unfoldConstants t
		x <- fresh (clName f)
		b <- valueToTerm $ f `clApp` (VId x w)
		return $ Lam x b ()
	VSort (VSUniv u)	-> return $ Star u
	VSort VSProp		-> return Prop
	VTop _			-> failure cannotReconstructTopError
	VApp v w		->
	    App <$> valueToTerm v <*> valueToTerm w

