
module TypeChecker.Monad.Datatypes where

import Control.Monad.State
import qualified Data.Map as Map

import Syntax.Internal
import Syntax.Values
import TypeChecker.Monad
import TypeChecker.Errors
import TypeChecker.Monad.Errors
import TypeChecker.UnfoldConstants
import Utils.Monads

getDatatypes :: TC Datatypes
getDatatypes = tcDatatypes <$> get

lookupDatatype :: Name -> TC DataInfo
lookupDatatype x =
    do	mcs <- Map.lookup x <$> getDatatypes
	case mcs of
	    Just cs -> return cs
	    Nothing -> failure $ becauseNotADatatype (vDef x $ error "lookupDatatype")

addDatatype :: Name -> DataInfo -> TC ()
addDatatype x cs =
    modify $ \s -> s { tcDatatypes = Map.insert x cs (tcDatatypes s) }

delDatatype :: Name -> TC ()
delDatatype x =
    modify $ \s -> s { tcDatatypes = Map.delete x (tcDatatypes s) }

setDatatypes :: Datatypes -> TC ()
setDatatypes ds =
    modify $ \s -> s { tcDatatypes = ds }

getDatatype :: Value -> TC (Name, [Value], [Value], [Name])
getDatatype v =
    do	v' <- unfoldConstants v
	case vAppView v' of
	    VAVDef x _ vs ->
		do  di <- lookupDatatype x
		    let cs = dataConstructors di
			(ps,is) = splitAt (dataParameters di) vs
		    return (x,ps,is,map fst cs)
	    _	-> failure $ becauseNotADatatype v

